#
#     MiAuth - Authenticate and interact with Xiaomi devices over BLE
#     Copyright (C) 2021  Daljeet Nandha
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
def int_to_bytes(i, size=4):
    res = bytearray(size)
    for j in range(size):
        res[j] = (i >> (j * 8)) & 0xff
    return res


def crc16(arr):
    n = ~sum(arr)
    return int_to_bytes(n, 2)



#
#     MiAuth - Authenticate and interact with Xiaomi devices over BLE
#     Copyright (C) 2021  Daljeet Nandha
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

#     This class is my Python port of https://github.com/scooterhacking/NinebotCrypto
#     Huge thanks to the original authors for sharing their work!

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.hashes import Hash, SHA1

from miauth.util import crc16


class NbCrypto(object):
    FW_DATA = bytes.fromhex("97CFB802844143DE56002B3B34780A5D")

    def __init__(self):
        self.name = b''
        self.sha1_key = b''
        self.ble_data = None
        self.app_data = None
        self.it = 0

    def set_name(self, name):
        self.name = name
        self.sha1_key = NbCrypto.calc_sha1_key(self.name, NbCrypto.FW_DATA)

    def set_ble_data(self, ble_data):
        self.ble_data = ble_data
        self.sha1_key = NbCrypto.calc_sha1_key(self.name, self.ble_data)

    def set_app_data(self, app_data):
        self.app_data = app_data
        self.sha1_key = NbCrypto.calc_sha1_key(self.app_data, self.ble_data)

    def gen_aes_data(self):
        aes_data = bytearray(16)
        aes_data[0] = 1
        aes_data[1] = (self.it & 0xff000000) >> 24
        aes_data[2] = (self.it & 0x00ff0000) >> 16
        aes_data[3] = (self.it & 0x0000ff00) >> 8
        aes_data[4] = (self.it & 0x000000ff) >> 0
        aes_data[5:5 + 8] = self.ble_data[:8]
        aes_data[15] = 0
        return aes_data

    def encrypt(self, data):
        result = bytearray(152)
        result[:3] = data[:3]

        pl_len = len(data) - 3
        pl = bytearray(pl_len)
        pl[:] = data[3:3 + pl_len]

        if self.it == 0 or self.ble_data is None:
            crc = NbCrypto.crc_next(pl)
            enc = NbCrypto.crypto_next(pl, self.sha1_key)
            result[3:3 + pl_len] = enc[:]
            result[pl_len + 3] = 0
            result[pl_len + 4] = 0
            result[pl_len + 5] = crc[0]
            result[pl_len + 6] = crc[1]
            result[pl_len + 7] = 0
            result[pl_len + 8] = 0
            result = result[:pl_len + 9]
        else:
            self.it += 1

            aes_data = self.gen_aes_data()
            enc = NbCrypto.crypto_next(pl, self.sha1_key, aes_data)

            aes_data[0] = 0x59
            aes_data[15] = pl_len
            crc = NbCrypto.crc_next(data, self.sha1_key, aes_data)

            result[3:3 + pl_len] = enc[:]
            result[pl_len + 3] = crc[0]
            result[pl_len + 4] = crc[1]
            result[pl_len + 5] = crc[2]
            result[pl_len + 6] = crc[3]
            result[pl_len + 7] = (self.it & 0xff00) >> 8
            result[pl_len + 8] = (self.it & 0x00ff) >> 0
            result = result[:pl_len + 9]

        return result

    def decrypt(self, data):
        result = bytearray(len(data) - 6)
        result[:3] = data[:3]

        pl_len = len(data) - 9
        pl = bytearray(pl_len)
        pl[:] = data[3:3 + pl_len]

        # update it from response (last two bytes)
        self.it = (data[-2] << 8) + data[-1]

        if self.it == 0 or self.ble_data is None:
            dec = NbCrypto.crypto_next(pl, self.sha1_key)
            result[3:] = dec[:]
        else:
            aes_data = self.gen_aes_data()
            dec = NbCrypto.crypto_next(pl, self.sha1_key, aes_data)
            result[3:] = dec[:]

        return result

    @classmethod
    def crc_next(cls, data, sha1_key=None, aes_data=None):
        if sha1_key is None and aes_data is None:  # crc_first
            return crc16(data)

        aes_key = cls.aes_ecb_encrypt(aes_data, sha1_key)

        xor_data1 = bytearray(16)
        xor_data1[:3] = data[:3]
        xor_data2 = bytearray(16)
        xor_data2[:] = aes_key[:]

        xor_data = cls.xor(xor_data1, xor_data2, 16)
        aes_key = cls.aes_ecb_encrypt(xor_data, sha1_key)
        xor_data2[:] = aes_key[:]

        pl_len = len(data) - 3
        byte_idx = 3
        while pl_len > 0:
            tmp_len = pl_len if pl_len <= 16 else 16

            xor_data1 = bytearray(16)
            xor_data1[:tmp_len] = data[byte_idx:byte_idx + tmp_len]

            # like in crypto_next, but first xor then aes
            xor_data = cls.xor(xor_data1, xor_data2, 16)

            aes_key = cls.aes_ecb_encrypt(xor_data, sha1_key)
            xor_data2[:] = aes_key[:]

            pl_len -= tmp_len
            byte_idx += tmp_len

        # TODO: this is not nice
        aes_data[0] = 1
        aes_data[15] = 0

        aes_key = cls.aes_ecb_encrypt(aes_data, sha1_key)
        xor_data1[:4] = aes_key[:4]

        crc = cls.xor(xor_data1, xor_data2, 4)

        return crc

    @classmethod
    def crypto_next(cls, inp_data, sha1_key, aes_data=None):
        result = bytearray(len(inp_data))

        byte_idx = 0
        pl_len = len(inp_data)

        while pl_len > 0:
            tmp_len = pl_len if pl_len <= 16 else 16
            xor_data1 = bytearray(16)
            xor_data1[:tmp_len] = inp_data[byte_idx:byte_idx + tmp_len]

            if aes_data is None:
                aes_key = cls.aes_ecb_encrypt(cls.FW_DATA, sha1_key)
            else:
                aes_data[15] += 1
                aes_key = cls.aes_ecb_encrypt(aes_data, sha1_key)

            xor_data2 = bytearray(16)
            xor_data2[:] = aes_key[:]

            xor_data = cls.xor(xor_data1, xor_data2, 16)
            result[byte_idx:byte_idx + tmp_len] = xor_data[:tmp_len]

            pl_len -= tmp_len
            byte_idx += tmp_len

        return result

    @staticmethod
    def calc_sha1_key(b1, b2):
        data = bytearray(32)
        data[:16] = b1
        data[16:] = b2

        sha = Hash(SHA1())
        sha.update(data)
        h = sha.finalize()[:16]
        assert len(h) == 16
        return h

    @staticmethod
    def aes_ecb_encrypt(data, key):
        aes = Cipher(
            algorithms.AES(key),
            modes.ECB(),
        ).encryptor()
        ct = aes.update(data) + aes.finalize()
        return ct

    @staticmethod
    def xor(d1, d2, size):
        result = bytearray(size)
        for i in range(size):
            result[i] = d1[i] ^ d2[i]
        return result



#
#     MiAuth - Authenticate and interact with Xiaomi devices over BLE
#     Copyright (C) 2021  Daljeet Nandha
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
class NbCommand(object):
    CMD_INIT = bytes.fromhex("5AA5003D215B00")
    CMD_PING = lambda rand: bytes.fromhex("5AA5103D215C00" + rand.hex())
    CMD_PAIR = lambda serial: bytes.fromhex("5AA50E3D215D00" + serial.hex())

    ACK_INIT = bytes.fromhex("5AA51E213D5B01")
    ACK_PRE = bytes.fromhex("5AA500213D5C00")
    ACK_PING = bytes.fromhex("5AA500213D5C01")
    ACK_PAIR = bytes.fromhex("5AA500213D5D01")
    ACK_LEN = len(ACK_INIT)



#
#     MiAuth - Authenticate and interact with Xiaomi devices over BLE
#     Copyright (C) 2022  Daljeet Nandha
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from miauth.ble.base import BLEBase
from miauth.ble.uuid import UUID

from bluepy import btle


class BluePy(BLEBase, btle.DefaultDelegate):
    def __init__(self, mac):
        self.p = btle.Peripheral()
        self.mac = mac

        self.handler = None

        self.channels = {
            UUID.AVDTP: None,
            UUID.UPNP: None,
            UUID.TX: None,
            UUID.RX: None,
            UUID.KEY: None,
        }

        btle.DefaultDelegate.__init__(self)
        BLEBase.__init__(self)

    def handleNotification(self, handle, data):
        if self.handler is None:
            raise Exception("BLE set handler first")
        if not data:
            return

        self.handler(data)

    def set_handler(self, handler):
        self.handler = handler

    def enable_notify(self, ch):
        val = b'\x01\x00'
        resp = self.p.writeCharacteristic(ch.valHandle + 1, val, True)
        if resp['rsp'][0] != 'wr':
            raise Exception("BLE could not setup notifications.")

    def disable_notify(self, ch):
        val = b'\x00\x00'
        resp = self.p.writeCharacteristic(ch.valHandle + 1, val, True)
        if resp['rsp'][0] != 'wr':
            raise Exception("BLE could not setup notifications.")

    def write(self, ch, data, resp=False):
        self.channels[ch].write(data, resp)

    def write_chunked(self, ch, data, resp=False, chunk_size=20):
        for i in range(0, len(data), chunk_size):
            chunk = data[i:i + chunk_size]
            self.write(ch, chunk, resp=resp)

    def write_parcel(self, ch, data, resp=False, chunk_size=18):
        for i in range(0, len(data), chunk_size):
            chunk = data[i:i + chunk_size]
            chunk = bytes([i // chunk_size + 1, 0]) + chunk
            self.write(ch, chunk, resp=resp)

    def connect(self):
        self.p.connect(self.mac, btle.ADDR_TYPE_RANDOM)
        self.p.setDelegate(self)

        for ch in self.p.getCharacteristics():
            uuid = UUID.from_hex(ch.uuid.binVal.hex())
            self.channels[uuid] = ch
            if "NOTIFY" in ch.propertiesToString():
                print("enabling notifications for:", uuid)
                self.enable_notify(ch)

    def disconnect(self):
        self.disable_notify(self.channels[UUID.RX])
        self.p.disconnect()

    def wait_notify(self, secs=1.0):
        while self.p.waitForNotifications(secs):
            continue

    def read_device_name(self):
        ch = self.p.getCharacteristics(uuid=btle.AssignedNumbers.deviceName)

        if not ch:
            raise Exception("Device name not found.")

        return ch[0].read()

    def has_channel(self, name):
        return name in self.channels and self.channels[name] is not None



#
#     MiAuth - Authenticate and interact with Xiaomi devices over BLE
#     Copyright (C) 2021  Daljeet Nandha
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

#     Huge thanks to Aaron Christophel (Atc1441) and Daniel Kucera!
#     I have adapted some function and variable declarations from their respective works:
#           https://github.com/danielkucera/mi-standardauth/blob/master/provision.py
#           https://github.com/atc1441/atc1441.github.io/blob/master/TelinkFlasher.html

import secrets

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat, PrivateFormat, NoEncryption
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.ciphers.aead import AESCCM
from cryptography.hazmat.primitives.hmac import HMAC
from cryptography.hazmat.primitives.kdf.hkdf import HKDF

from miauth.util import crc16, int_to_bytes


class MiCrypto(object):
    @staticmethod
    def gen_rand_key():
        return secrets.token_bytes(16)

    @staticmethod
    def bytes_to_pub_key(data):
        return ec.EllipticCurvePublicKey.from_encoded_point(
            ec.SECP256R1(), b'\x04' + data)

    @staticmethod
    def pub_key_to_bytes(key):
        return key.public_bytes(Encoding.X962, PublicFormat.UncompressedPoint)[1:]

    @staticmethod
    def gen_keypair():
        priv_key = ec.generate_private_key(ec.SECP256R1(), default_backend())
        pub_key = priv_key.public_key()
        return priv_key, pub_key

    @staticmethod
    def val_to_private_key(val):
        return ec.derive_private_key(val, ec.SECP256R1())

    @staticmethod
    def private_key_to_val(key):
        return key.private_numbers().private_value

    @staticmethod
    def generate_secret(private_key, public_key):
        return private_key.exchange(ec.ECDH(), public_key)

    @staticmethod
    def derive_key(shared_key, salt=None):
        info = b"mible-login-info"
        if salt is None:
            info = b"mible-setup-info"

        return HKDF(
            algorithm=hashes.SHA256(),
            length=64,
            salt=salt,
            info=info,
            backend=default_backend()
        ).derive(shared_key)

    @staticmethod
    def hash(derived_key, data):
        hmac = HMAC(derived_key, algorithm=hashes.SHA256())
        hmac.update(data)
        return hmac.finalize()

    @staticmethod
    def encrypt_did(key, did):
        aes_ccm = AESCCM(key, tag_length=4)
        nonce = bytes([16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27])
        did = did
        aad = b"devID"

        return aes_ccm.encrypt(nonce, did, aad)

    @staticmethod
    def encrypt_uart(key, iv, msg, it=0, rand=secrets.token_bytes(4)):
        msg = msg[2:]  # ditch header

        size = msg[:1]
        data = msg[1:]
        data += rand  # add four random bytes to data

        it = int_to_bytes(it)  # encode iterator to four bytes
        nonce = iv + bytes([0] * 4) + it

        aes_ccm = AESCCM(key, tag_length=4)
        ct = aes_ccm.encrypt(nonce, data, None)

        header = b'\x55\xab'  # new header
        data = size + it[:2] + ct  # new data
        crc = crc16(data)  # new checksum

        return header + data + crc

    @staticmethod
    def decrypt_uart(key, iv, msg):
        header = msg[:2]
        if header != b'\x55\xab':
            raise Exception("Invalid header.")

        it = msg[3:5]
        ct = msg[5:-2]

        nonce = iv + bytes([0] * 4) + it + bytes([0] * 2)

        aes_ccm = AESCCM(key, tag_length=4)
        return aes_ccm.decrypt(nonce, ct, None)



#
#     MiAuth - Authenticate and interact with Xiaomi devices over BLE
#     Copyright (C) 2021-2022  Daljeet Nandha
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import time

from miauth.mi.micommand import MiCommand
from miauth.mi.micrypto import MiCrypto
from miauth.ble.base import BLEBase
from miauth.ble.uuid import UUID
from miauth.util import crc16


class MiClient(object):
    class State(object):
        INIT = -1
        RECV_INFO = 0
        SEND_KEY = 1
        RECV_KEY = 2
        SEND_DID = 3
        CONFIRM = 4
        COMM = 5

    def __init__(self, ble: BLEBase, debug=False):
        self.ble = ble
        self.debug = debug

        self.ble.set_handler(self.main_handler)

        # TODO: implement power button press recognition self.button = False
        # self.s = btle.Scanner().withDelegate(self)

        # state machine is supplied with a sequence of ...
        # if seq = [<state>, <on_state_enter_func()>]
        # TODO: create sequence controller class
        self.seq = ()
        self.seq_idx = 0

        # buffer for send handler
        self.send_data = b''

        # buffer for receive handler
        self.receive_frames = 0
        self.received_data = b''

        # authentication related stuff
        self.remote_info = b''
        self.remote_key = b''
        self.token = b''
        self.keys = b''

        # counter for sent uart commands
        self.uart_it = 0

    def main_handler(self, data):
        if len(data) == 0:
            return

        frm = data[0] + 0x100 * data[1]
        if self.debug:
            print("<-", data.hex(), self.get_state())

        frm = data[0]
        if len(data) > 1:
            frm += 0x100 * data[1]

        if self.get_state() in [MiClient.State.RECV_INFO,
                                MiClient.State.RECV_KEY]:
            self.receive_handler(frm, data)
        elif self.get_state() in [MiClient.State.SEND_KEY,
                                  MiClient.State.SEND_DID]:
            self.send_handler(frm, data)
        elif self.get_state() == MiClient.State.CONFIRM:
            self.confirm_handler(frm)
        elif self.get_state() == MiClient.State.COMM:
            # TODO: check if correct number of frames received
            self.received_data += data

    def connect(self):
        self.ble.connect()

    def disconnect(self):
        self.ble.disconnect()

    def reset(self):
        self.ble.disconnect()
        self.__init__(self.ble, debug=self.debug)

    def get_state(self):
        return self.seq[self.seq_idx][0]

    def next_state(self):
        self.seq_idx += 1

        if self.debug:
            print("new state:", self.get_state())

        # exec entry func
        f = self.seq[self.seq_idx][1]
        if f is not None:
            f()

    def receive_handler(self, frm, data):
        if frm == 0:
            self.receive_frames = data[4] + 0x100 * data[5]
            if self.debug:
                print("Expecting", self.receive_frames, "frames")

            self.received_data = b''
            self.ble.write(UUID.AVDTP, MiCommand.RCV_RDY)
        else:
            self.received_data += data[2:]

        if frm == self.receive_frames:
            if self.debug:
                print("All frames received: ", self.received_data.hex())
            self.ble.write(UUID.AVDTP, MiCommand.RCV_OK)
            self.next_state()

    def send_handler(self, frm, data):
        if frm != 0:
            return

        if data == MiCommand.RCV_RDY:
            if self.debug:
                print("Mi ready to receive key")
            self.ble.write_parcel(UUID.AVDTP, self.send_data)
        if data == MiCommand.RCV_TOUT:
            raise Exception("Mi sent RCV timeout.")
        if data == MiCommand.RCV_ERR:
            raise Exception("Mi sent some RCV error?")
        if data == MiCommand.RCV_OK:
            if self.debug:
                print("Mi confirmed key receive")
            self.next_state()

    def confirm_handler(self, frm):
        if frm == 0x11:
            print("Mi authentication successful!")
        elif frm == 0x12:
            print("Mi authentication failed!")
        elif frm == 0x21:
            print("Mi login successful!")
        elif frm == 0x23:
            print("Mi login failed!")
        else:
            print("Mi unknown response...")
        self.next_state()

    def calc_did(self, private_key):
        remote_pub_key = MiCrypto.bytes_to_pub_key(self.remote_key)
        e_share_key = MiCrypto.generate_secret(private_key, remote_pub_key)

        derived_key = MiCrypto.derive_key(e_share_key)

        token = derived_key[0:12]
        bind_key = derived_key[12:28]
        a = derived_key[28:44]

        did_ct = MiCrypto.encrypt_did(a, self.remote_info[4:])

        if self.debug:
            print("eShareKey:", e_share_key.hex())
            print("HKDF result: ", derived_key.hex())
            print("token:", token.hex())
            print("bind_key:", bind_key.hex())
            print("A:", a.hex())
            print("AES did CT: ", did_ct.hex())

        return did_ct, token

    def calc_login_info(self, random_key):
        salt = random_key + self.remote_key
        salt_inv = self.remote_key + random_key

        derived_key = MiCrypto.derive_key(self.token, salt=salt)
        keys = {
            'dev_key': derived_key[:16],
            'app_key': derived_key[16:32],
            'dev_iv': derived_key[32:36],
            'app_iv': derived_key[36:40],
        }
        info = MiCrypto.hash(keys['app_key'], salt)
        expected_remote_info = MiCrypto.hash(keys['dev_key'], salt_inv)

        if self.debug:
            print("HKDF result:", derived_key.hex())
            for key, val in keys.items():
                print(f"{key.upper()}:", val.hex())

        return info, expected_remote_info, keys

    def register(self):
        priv_key, pub_key = MiCrypto.gen_keypair()
        if self.debug:
            print("Private Key (Val):", MiCrypto.private_key_to_val(priv_key))
            print("Public Key (Hex):", MiCrypto.pub_key_to_bytes(pub_key).hex())

        def on_recv_info_state():
            self.ble.write(UUID.UPNP, MiCommand.CMD_GET_INFO)

        def on_send_key_state():
            self.remote_info = self.received_data

            self.send_data = MiCrypto.pub_key_to_bytes(pub_key)
            self.ble.write(UUID.UPNP, MiCommand.CMD_SET_KEY)
            self.ble.write(UUID.AVDTP, MiCommand.CMD_SEND_DATA)

        def on_send_did_state():
            self.remote_key = self.received_data

            self.send_data, self.token = self.calc_did(priv_key)
            self.ble.write(UUID.AVDTP, MiCommand.CMD_SEND_DID)

        def on_confirm_state():
            self.ble.write(UUID.UPNP, MiCommand.CMD_AUTH)

        self.seq = ((MiClient.State.INIT, None),
                    (MiClient.State.RECV_INFO, on_recv_info_state()),
                    (MiClient.State.SEND_KEY, on_send_key_state),
                    (MiClient.State.RECV_KEY, None),
                    (MiClient.State.SEND_DID, on_send_did_state),
                    (MiClient.State.CONFIRM, on_confirm_state),
                    (MiClient.State.COMM, None),
                    )
        self.seq_idx = 0

        self.next_state()
        while self.get_state() != MiClient.State.COMM:
            self.ble.wait_notify(secs=3.0)

            if self.get_state() != MiClient.State.COMM:
                # Trick 17: if no response ...
                # disconnect here and wait for power button press
                # after button press, reconnect and restart from beginning
                self.reset()

                print(">> Please press power button within 5 secs after beep")
                time.sleep(5)
                self.ble.connect()

                return self.register()  # return because of recursion
            else:
                break

    def save_token(self, filename):
        with open(filename, 'wb') as f:
            f.write(self.token)

    def load_token(self, filename):
        with open(filename, 'rb') as f:
            self.token = f.read()

    def login(self):
        rand_key = MiCrypto.gen_rand_key()

        def on_send_key_state():
            self.send_data = rand_key
            self.ble.write(UUID.UPNP, MiCommand.CMD_LOGIN)
            self.ble.write(UUID.AVDTP, MiCommand.CMD_SEND_KEY)

        def on_recv_info_state():
            self.remote_key = self.received_data

        def on_send_did_state():
            self.remote_info = self.received_data

            self.send_data, expected_remote_info, self.keys = self.calc_login_info(rand_key)
            assert self.remote_info == expected_remote_info, \
                f"{self.remote_info.hex()} != {expected_remote_info.hex()}"
            self.ble.write(UUID.AVDTP, MiCommand.CMD_SEND_INFO)

        self.seq = (
            (MiClient.State.INIT, None),
            (MiClient.State.SEND_KEY, on_send_key_state),
            (MiClient.State.RECV_KEY, None),
            (MiClient.State.RECV_INFO, on_recv_info_state),
            (MiClient.State.SEND_DID, on_send_did_state),
            (MiClient.State.CONFIRM, None),
            (MiClient.State.COMM, None),
        )
        self.seq_idx = 0

        self.next_state()
        while self.get_state() != MiClient.State.COMM:
            self.ble.wait_notify(secs=3.0)

    def comm(self, cmd):
        if self.get_state() != MiClient.State.COMM:
            raise Exception("Not in COMM state. Retry maybe.")

        if type(cmd) not in [bytearray, bytes]:
            cmd = bytes.fromhex(cmd)

        if cmd[:2] != b'\x55\xAA':
            if cmd[:2] == b'\x5a\xa5':
                raise Exception("Command must start with 55 AA (M365 PROTOCOl)!\
                                You sent a Nb command, try Nb pairing instead.")
            else:
                raise Exception("Command must start with 55 AA (M365 PROTOCOl)!")

        self.received_data = b''
        if not self.keys:
            self.ble.write_chunked(UUID.TX, cmd)

            self.ble.wait_notify()

            if not self.received_data:
                raise Exception("No answer received. Try login first.")

            return self.received_data

        res = MiCrypto.encrypt_uart(self.keys['app_key'],
                                    self.keys['app_iv'],
                                    cmd,
                                    it=self.uart_it)
        self.ble.write_chunked(UUID.TX, res)
        self.uart_it += 1

        self.ble.wait_notify()

        if not self.received_data:
            #raise Exception("No answer received. Firmware not supported.")
            print("No answer received")
            return

        return MiCrypto.decrypt_uart(
            self.keys['dev_key'],
            self.keys['dev_iv'],
            self.received_data)[3:-4]



#
#     MiAuth - Authenticate and interact with Xiaomi devices over BLE
#     Copyright (C) 2022  Daljeet Nandha
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from abc import ABC, abstractmethod


class BLEBase(ABC):

    @abstractmethod
    def set_handler(self, handler):
        raise NotImplementedError

    @abstractmethod
    def enable_notify(self, uuid):
        raise NotImplementedError

    @abstractmethod
    def write(self, ch, data, resp=False):
        raise NotImplementedError

    @abstractmethod
    def write_chunked(self, ch, data, resp=False, chunk_size=20):
        raise NotImplementedError

    @abstractmethod
    def write_parcel(self, ch, data, resp=False, chunk_size=18):
        raise NotImplementedError

    @abstractmethod
    def connect(self):
        raise NotImplementedError

    @abstractmethod
    def disconnect(self):
        raise NotImplementedError

    @abstractmethod
    def wait_notify(self, secs=1.0):
        raise NotImplementedError

    @abstractmethod
    def read_device_name(self):
        raise NotImplementedError

    @abstractmethod
    def has_channel(self):
        raise NotImplementedError



#
#     MiAuth - Authenticate and interact with Xiaomi devices over BLE
#     Copyright (C) 2021  Daljeet Nandha
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
class MiCommand(object):
    CMD_GET_INFO = b"\xa2\x00\x00\x00"
    CMD_SET_KEY = b"\x15\x00\x00\x00"
    CMD_LOGIN = b"\x24\x00\x00\x00"
    CMD_AUTH = b"\x13\x00\x00\x00"

    CMD_SEND_DATA = b"\x00\x00\x00\x03\x04\x00"
    CMD_SEND_DID = b"\x00\x00\x00\x00\x02\x00"
    CMD_SEND_KEY = b"\x00\x00\x00\x0b\x01\x00"
    CMD_SEND_INFO = b"\x00\x00\x00\x0a\x02\x00"

    RCV_RDY = b"\x00\x00\x01\x01"
    RCV_OK = b"\x00\x00\x01\x00"
    RCV_TOUT = b"\x00\x00\x01\x05\x01\x00"
    RCV_ERR = b"\x00\x00\x01\x05\x03\x00"
    # RCV_RESP_KEY = b"\x00\x00\x00\x0d\x01\x00"
    # RCV_RESP_INFO = b"\x00\x00\x00\x0c\x02\x00"




#
#     MiAuth - Authenticate and interact with Xiaomi devices over BLE
#     Copyright (C) 2021  Daljeet Nandha
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#
#     Huge thanks and credits to Camilo (CamiAlfa) for his work on 55aa:
#       https://github.com/CamiAlfa/M365-BLE-PROTOCOL
#       https://github.com/CamiAlfa/M365-BLE-PROTOCOL/blob/master/protocolo
#     Huge thanks and credits to Piotr Dobrowolski, adapted 55ab from this work:
#       https://github.com/Informatic/py9b
#       https://github.com/Informatic/py9b/issues/2#issuecomment-549683873
#       https://github.com/CamiAlfa/M365-BLE-PROTOCOL/issues/7#issuecomment-546872603

from bluepy import btle

from miauth.mi.micommand import MiCommand
from miauth.ble.uuid import UUID
from miauth.ble.base import BLEBase
from miauth.util import crc16


class M365Client(btle.DefaultDelegate):
    def __init__(self, ble: BLEBase, debug=False):
        self.ble = ble
        self.debug = debug

        self.ble.set_handler(self.main_handler)

        self.ch_tx = None
        self.ch_rx = None

        self.ch_key = None
        self.key = None

        # buffer for receive handler
        self.receive_frames = 0
        self.received_data = b''

        # counter for sent uart commands
        self.uart_it = 0

    def main_handler(self, data):
        if len(data) == 0:
            return
        self.received_data += data

    def connect(self):
        self.ble.connect()

    def disconnect(self):
        self.ble.disconnect()

    def reset(self):
        self.ble.disconnect()
        self.__init__(self.p, self.mac, debug=self.debug)

    def recover_key(self):
        if self.ble.has_channel(UUID.KEY):
            print("found old 55ab encryption -> recovering key")
            self.key = self.ch_key[0].read()
            self.key += self.comm("55aa0322015020")[9:]
            if self.debug:
                print("key:", self.key.hex())

    def comm(self, cmd):
        if type(cmd) not in [bytearray, bytes]:
            cmd = bytes.fromhex(cmd)

        if cmd[:2] != b'\x55\xaa':
            if cmd[:2] == b'\x5a\xa5':
                raise Exception("Command must start with 55 AA (M365 PROTOCOl)!\
                                You sent a Nb command, try Nb pairing instead.")
            else:
                raise Exception("Command must start with 55 AA (M365 PROTOCOl)!")

        self.received_data = b''

        if self.key:
            len_ = cmd[2:3]
            cmd = b'\x55\xab' + len_ + self.crypt(cmd[3:] + b'\x00\x00\x00\x00')
        res = cmd + crc16(cmd[2:])  # new checksum
        self.ble.write_chunked(UUID.TX, res)
        self.uart_it += 1

        self.ble.wait_notify()

        if not self.received_data:
            raise Exception("No answer received. Firmware not supported.")

        if self.debug:
            print("received:", self.received_data.hex())
        res = self.received_data
        if crc16(res[2:-2]) != res[-2:]:
            raise Exception("Checksum mismatch in response")

        if self.key:
            res = self.crypt(res[3:])[:-4]

        if self.debug:
            print("response:", res.hex())
        return res[3:-2]

    def crypt(self, data):
        k = self.key
        return bytearray([b ^ (k[i] if i < len(k) else 0) for i, b in enumerate(data)])



#
#     MiAuth - Authenticate and interact with Xiaomi devices over BLE
#     Copyright (C) 2021-2022  Daljeet Nandha
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
class UUID(object):
    UART = "6e400001-b5a3-f393-e0a9-e50e24dcca9e"
    TX = "6e400002-b5a3-f393-e0a9-e50e24dcca9e"
    RX = "6e400003-b5a3-f393-e0a9-e50e24dcca9e"
    AUTH = "0000fe95-0000-1000-8000-00805f9b34fb"
    UPNP = "00000010-0000-1000-8000-00805f9b34fb"
    AVDTP = "00000019-0000-1000-8000-00805f9b34fb"
    KEY = "00000014-0000-1000-8000-00805f9b34fb"

    @staticmethod
    def from_hex(val):
        if val.lower() == UUID.UART.replace("-", ""):
            return UUID.UART
        elif val.lower() == UUID.TX.replace("-", ""):
            return UUID.TX
        elif val.lower() == UUID.RX.replace("-", ""):
            return UUID.RX
        elif val.lower() == UUID.AUTH.replace("-", ""):
            return UUID.AUTH
        elif val.lower() == UUID.UPNP.replace("-", ""):
            return UUID.UPNP
        elif val.lower() == UUID.AVDTP.replace("-", ""):
            return UUID.AVDTP
        elif val.lower() == UUID.KEY.replace("-", ""):
            return UUID.KEY



#
#     MiAuth - Authenticate and interact with Xiaomi devices over BLE
#     Copyright (C) 2021  Daljeet Nandha
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from unittest import TestCase

from miauth.util import crc16, int_to_bytes


class Test(TestCase):
    def test_int_to_bytes(self):
        b = int_to_bytes(1337)
        self.assertEqual(bytes([0x39, 5, 0, 0]), b, b.hex())

    def test_crc16(self):
        crc = crc16(bytes([0xa1, 0x21, 0xf3, 4, 5, 6, 7, 8, 9]))
        self.assertEqual(bytes.fromhex("23fe"), crc, crc.hex())



#
#     MiAuth - Authenticate and interact with Xiaomi devices over BLE
#     Copyright (C) 2021-2022  Daljeet Nandha
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import time
import threading
from collections import deque
from enum import Enum
import secrets

from miauth.nb.nbcommand import NbCommand
from miauth.ble.uuid import UUID
from miauth.ble.base import BLEBase


class NbClient(object):
    class State(Enum):
        DISC = 0
        CON = 1
        PING = 2
        PRE = 3
        PAIR = 4
        COMM = 5

    APP_KEY = secrets.token_bytes(16)

    def __init__(self, ble: BLEBase, crypto, debug=False):
        self.ble = ble
        self.crypto = crypto
        self.debug = debug

        self.ble.set_handler(self.main_handler)

        self.received_serial = b''
        self.received_key = b''
        self.receive_buffer = b''
        self.send_buffer = deque()

        self.state = NbClient.State.DISC

    def main_handler(self, data):
        if data[:2] == b"\x5a\xa5":
            self.receive_buffer = data
        else:
            self.receive_buffer += data

        dec = self.crypto.decrypt(self.receive_buffer)
        if self.debug:
            print("Received message:", self.receive_buffer.hex())
            print("Decoded message:", dec.hex())

        if len(dec) == self.receive_buffer[2] + 7:
            cmd = dec[:NbCommand.ACK_LEN]
            payload = dec[NbCommand.ACK_LEN:]
            self.receive_handler(cmd, payload)

    def connect(self):
        self.ble.connect()
        self.state = NbClient.State.CON

    def disconnect(self):
        self.ble.disconnect()
        self.state = NbClient.State.DISC

    def receive_handler(self, cmd, payload):
        if self.debug:
            print("Got cmd:", cmd.hex())

        if cmd == NbCommand.ACK_INIT:
            self.received_key = payload[:16]
            self.received_serial = payload[16:]
            if self.debug:
                print("> BLE Key:", self.received_key.hex())
                print("> Serial:", self.received_serial.decode())
                print("Setting ble data/key in crypto")
            self.crypto.set_ble_data(self.received_key)

            self.state = NbClient.State.PING
        elif cmd == NbCommand.ACK_PRE:
            self.state = NbClient.State.PRE
        elif cmd == NbCommand.ACK_PING:
            if self.debug:
                print("Setting app data/key in crypto")

            self.crypto.set_app_data(NbClient.APP_KEY)

            self.state = NbClient.State.PAIR
        elif cmd == NbCommand.ACK_PAIR:
            print("Nb authentication successful!")

            self.state = NbClient.State.COMM
        else:
            self.receive_buffer = payload

    def process_thread(self):
        while self.state != NbClient.State.COMM:
            if self.debug:
                print("Current state:", self.state)

            if self.state == NbClient.State.CON:
                self.send_buffer.append(
                    self.crypto.encrypt(NbCommand.CMD_INIT)
                )
            elif self.state == NbClient.State.PING:
                self.send_buffer.append(
                    self.crypto.encrypt(NbCommand.CMD_PING(NbClient.APP_KEY))
                )
            elif self.state == NbClient.State.PRE:
                print(">> Please press the POWER button on the device")

                # Can send either Ping or Pair message, doesn't matter
                self.send_buffer.append(
                    self.crypto.encrypt(NbCommand.CMD_PAIR(self.received_serial))
                )
            elif self.state == NbClient.State.PAIR:
                self.send_buffer.append(
                    self.crypto.encrypt(NbCommand.CMD_PAIR(self.received_serial))
                )

            time.sleep(2.0)

    def send_thread(self):
        while self.state != NbClient.State.DISC:
            if not self.send_buffer:
                time.sleep(1.0)
                continue

            msg = self.send_buffer.pop()
            if self.debug:
                print("Sending message:", msg.hex())

            # frame has 32 - 12 = 20 bytes --> need to chunk!!
            msg_len = len(msg)
            byte_idx = 0
            while msg_len > 0:
                tmp_len = msg_len if msg_len <= 20 else 20
                buf = msg[byte_idx:byte_idx + tmp_len]
                self.ble.write(UUID.TX, buf)

                msg_len -= tmp_len
                byte_idx += tmp_len

            time.sleep(1.0)

    def auth(self):
        self.crypto.set_name(self.ble.read_device_name())

        threading.Thread(target=self.process_thread).start()
        threading.Thread(target=self.send_thread).start()

        while self.state != NbClient.State.COMM:
            time.sleep(1.0)

    def comm(self, cmd):
        if self.state != NbClient.State.COMM:
            raise Exception("Not in COMM state. Retry.")

        if type(cmd) not in [bytearray, bytes]:
            cmd = bytes.fromhex(cmd)

        self.receive_buffer = b''
        while not self.receive_buffer:
            self.send_buffer.append(self.crypto.encrypt(cmd))
            time.sleep(2.0)

        return self.receive_buffer



#
#     MiAuth - Authenticate and interact with Xiaomi devices over BLE
#     Copyright (C) 2021  Daljeet Nandha
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from unittest import TestCase

from miauth.mi.micrypto import MiCrypto


class TestMiCrypto(TestCase):
    def test_gen_rand_key(self):
        rand = MiCrypto.gen_rand_key()
        self.assertEqual(16, len(rand))

    def test_gen_keypair(self):
        _, pub_key = MiCrypto.gen_keypair()
        pub_key = MiCrypto.pub_key_to_bytes(pub_key)
        self.assertEqual(64, len(pub_key))

    def test_gen_private_key(self):
        val = 38598657185418289442228743809972412250162588503372669841026540598582897367118
        priv_key = MiCrypto.val_to_private_key(val)
        priv_val = MiCrypto.private_key_to_val(priv_key)
        pub = MiCrypto.pub_key_to_bytes(priv_key.public_key())
        self.assertEqual(val, priv_val)
        self.assertEqual(bytes.fromhex("b5dca0aec31a8932d0f53cbcbcf0cfdd833c355cada1025cc076e013439ddec2b4017b546a11d79a758db9d015a2ed8926cf82179b593679187d623b5e430fca"),
                         pub,
                         pub.hex())

    def test_create_e_share_key(self):
        kp1 = MiCrypto.gen_keypair()
        kp2 = MiCrypto.gen_keypair()
        secret1 = MiCrypto.generate_secret(kp1[0], kp2[1])
        secret2 = MiCrypto.generate_secret(kp2[0], kp1[1])
        self.assertEqual(secret1, secret2)
        self.assertEqual(32, len(secret1))

    def test_derive_key(self):
        secret = bytes.fromhex("5a3d987d45f6484aff82ffde1e9105b7f6cc79fa7467f12c5855ad9e3f1d8f2f")
        derived = MiCrypto.derive_key(secret, bytes([1, 2, 3, 4]))
        self.assertEqual(bytes.fromhex("40ccc0ee058c3a1d37c08e6f72bc2c57c0a406aaa801a0b1b72f22c8c3ec930d3f151e2eb38a2303d8625a18084daa15667496dcfbc53ba3074ce35d6c90d987"),
                         derived,
                         derived.hex())

        derived = MiCrypto.derive_key(secret)
        self.assertEqual(bytes.fromhex("104ec0eda032b6d213c245359e585d3bfd4b7c5d683c99f49fd86aaf0de0f6b0bfafb897e3b3727aaa8f8ad6b21a737c1d85c3aae340969f268d2d95ca8848c1"),
                         derived,
                         derived.hex())

    def test_hash(self):
        key = bytes.fromhex("E2B274F08128A62A9575288BED169B3E")
        hash = MiCrypto.hash(key, bytes([1, 2, 3, 4]))
        self.assertEqual(bytes.fromhex("235d7f910974acb594d76a1652a856ce4f269e3060d7c8512e94b2da345d3083"),
                         hash,
                         hash.hex())

    def test_encrypt_did(self):
        key = bytes.fromhex("4FEB7165982BF1C6183A51B8CADD0EEC")
        ct = MiCrypto.encrypt_did(key, bytes([1, 2, 3, 4]))
        self.assertEqual(bytes.fromhex("aeebd70f8c2bdf8c"), ct, ct.hex())

    def test_encrypt_uart(self):
        app_key = bytes.fromhex("239b3c7e92dc6d6d2fa174a215aedf2e")
        app_iv = bytes([1, 2])
        ct = MiCrypto.encrypt_uart(app_key, app_iv, bytes.fromhex("55aa032001100e"), rand=bytes([1, 2, 3, 4]))
        self.assertEqual(bytes.fromhex("55ab030000adf399086b9e0bd059366ad10dfa"), ct, ct.hex())

    def test_decrypt_uart(self):
        dev_key = bytes.fromhex("239b3c7e92dc6d6d2fa174a215aedf2e")
        dev_iv = bytes([1, 2])
        ct = MiCrypto.encrypt_uart(dev_key, dev_iv, bytes.fromhex("55aa032001100e"), rand=bytes([1, 2, 3, 4]))
        self.assertEqual(bytes.fromhex("55ab030000adf399086b9e0bd059366ad10dfa"), ct, ct.hex())
        
    def test_register(self):
        priv_key = MiCrypto.val_to_private_key(48461508383982493215332654270464913273532832436436077476553357014100094140803)
        #print(MiCrypto.pub_key_to_bytes(priv_key.public_key()).hex())
        
        remote_info = bytes.fromhex("0100000000626c742e342e31386e35383236366b67673030")
        remote_pub_key = MiCrypto.bytes_to_pub_key(bytes.fromhex("2afe2a8c1c56e5e70721665cd20d017273111ecaeceb1e4d641e7b7a122a9c3041e5cbc962eefbdb155ffd95847a0d8762803291fc2866c5672ceee0e77d77fc"))
        
        secret = MiCrypto.generate_secret(priv_key, remote_pub_key)
        derived = MiCrypto.derive_key(secret)
        did_key = derived[28:44]
        did_ct = MiCrypto.encrypt_did(did_key, remote_info[4:])
        self.assertEqual("fac3a6fd591dcea21f9f4fefe297804f49291527ae818b285f4a75a6fab72af8", secret.hex())
        self.assertEqual("0cf5615003810d89c233a12a8fc5100e31299d80c4c290dc7d33f19ec42ea48a95c5544f105fe7ebb8b39233c6542b1fff90b2206265080bf516365fd8d758fe", derived.hex())
        self.assertEqual("c42ea48a95c5544f105fe7ebb8b39233", did_key.hex())
        self.assertEqual("646735cc7a96373aabbd93afa089bb6cd2d080302101007a", did_ct.hex())

    def test_login(self):
        token = bytes.fromhex("0cf5615003810d89c233a12a")
        random_key = bytes.fromhex("a8699783c1f03c7b73a046cdb613a9bf")
        remote_key = bytes.fromhex("90fdec0ece05016d7f116b50fca4b4bf")
        
        remote_info = bytes.fromhex("471467ea7ed6064f8dd72f416c079dcb3bb78e3c94a51e97b98ef7623a7798e5")

        salt = random_key + remote_key
        salt_inv = remote_key + random_key

        derived_key = MiCrypto.derive_key(token, salt=salt)
        keys = {
            'dev_key': derived_key[:16],
            'app_key': derived_key[16:32],
            'dev_iv': derived_key[32:36],
            'app_iv': derived_key[36:40],
        }
        info = MiCrypto.hash(keys['app_key'], salt)
        expected_remote_info = MiCrypto.hash(keys['dev_key'], salt_inv)

        self.assertEqual(remote_info, expected_remote_info)
        self.assertEqual("bbb99b6a6f1ae419e3b13db93514f3e12a1843033f276392eea99068affca753", info.hex())
        self.assertEqual("3c2fdae69f8746663a7d91029724b072", keys['dev_key'].hex())
        self.assertEqual("12a4122dbe3cf60f3bdcb37139c34611", keys['app_key'].hex())
        self.assertEqual("cd016050", keys['dev_iv'].hex())
        self.assertEqual("c87de6f1", keys['app_iv'].hex())



#
#     MiAuth - Authenticate and interact with Xiaomi devices over BLE
#     Copyright (C) 2021-2022  Daljeet Nandha
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import sys
import os
import argparse

from miauth.mi.miclient import MiClient
from miauth.mi.m365client import M365Client
from miauth.nb.nbclient import NbClient
from miauth.nb.nbcrypto import NbCrypto

from miauth.ble.blue import BluePy

parser = argparse.ArgumentParser()
parser.add_argument("mac", help="mac address of target device")
parser.add_argument("-m", "--m365", action='store_true', help="use M365 protocol instead")
parser.add_argument("-n", "--nb", action='store_true', help="use Nb protocol instead")
parser.add_argument("-c", "--command", help="send command (w/o checksum) to uart and print reply")
parser.add_argument("-s", "--serial", action='store_true', help="retrieve serial number")
parser.add_argument("-v", "--version", action='store_true', help="retrieve firmware version")
parser.add_argument("-d", "--debug", action='store_true', help="activate debug log")

parser.add_argument("-r", "--register", action='store_true',
                    help="register with device / create token (caution: will lose bond to all other apps)")
parser.add_argument("-t", "--token_file", default="./mi_token",
                    help="path to mi token file (default: ./mi_token)")

args = parser.parse_args()
print(args)


def nb_main(ble):
    nc = NbClient(ble, NbCrypto(), debug=args.debug)

    print("Connecting")
    nc.connect()

    print("Authenticating")
    nc.auth()

    if args.command:
        resp = nc.comm(args.command)
        print("UART reply:", resp.hex())

    # NOTE: don't send checksums!
    if args.serial:
        print("Retrieving serial number")
        resp = nc.comm("5aa5013d2001100e")
        print("Serial no.:", resp.decode())
    if args.version:
        print("Retrieving firmware version")
        resp = nc.comm("5aa5013d20011a10")
        print("Firmware version:", f"{resp[0]}.{resp[1]}")

    print("Disconnecting")
    nc.disconnect()


def m365_main(ble):
    mc = M365Client(ble, debug=args.debug)

    print("Connecting")
    mc.connect()
    mc.recover_key()

    if args.command:
        resp = mc.comm(args.command)
        print("UART reply:", resp.hex())

    # NOTE: don't send checksums!
    if args.serial:
        print("Retrieving serial number")
        resp = mc.comm("55aa032001100e")
        print("Serial no.:", resp.decode())
    if args.version:
        print("Retrieving firmware version")
        resp = mc.comm("55aa0320011a10")
        print("Firmware version:", f"{resp[0]}.{resp[1]}")

    print("Disconnecting")
    mc.disconnect()


def mi_main(ble):
    mc = MiClient(ble, debug=args.debug)

    print("Connecting")
    mc.connect()

    if args.register:
        print("Registering")
        mc.register()
        mc.save_token(args.token_file)
        print("Saved token to:", args.token_file)

    if not mc.token:
        if not os.path.isfile(args.token_file):
            sys.exit("""
No authentication token found, register with '-r' or specify path to token file with '-t <path>'.
Caution: After registration this device will lose coupling to all other apps (remove/add device in Mi Home app if needed).
                     """)

        print("Loading token from:", args.token_file)
        mc.load_token(args.token_file)

    print("Logging in...")
    mc.login()

    if args.command:
        print("Sending command:", args.command)
        resp = mc.comm(args.command)
        print("UART reply:", resp.hex())

    # NOTE: don't send checksum
    if args.serial:
        print("Retrieving serial number")
        resp = mc.comm("55aa032001100e")
        print("Serial number:", resp.decode())
    if args.version:
        print("Retrieving firmware version")
        resp = mc.comm("55aa0320011a10")
        print("Firmware version:", f"{resp[0]}.{resp[1]}")

    print("Disconnecting")
    mc.disconnect()


def main():
    # default BLE lib is bluepy
    ble = BluePy(args.mac)

    if args.nb:
        print("Using Nb")
        nb_main(ble)
    elif args.m365:
        m365_main(ble)
    else:
        print("Using Mi")
        mi_main(ble)


if __name__ == "__main__":
    main()



from setuptools import setup, find_packages

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='miauth',
    version='0.9.6',
    url='https://github.com/dnandha/miauth',
    license='GNU AGPL v3',
    author='Daljeet Nandha',
    author_email='dalj337@gmail.com',
    description='Authenticate and interact with Xiaomi devices over BLE',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(include=['miauth', 'miauth.*']),
    entry_points={
        'console_scripts': ['miauth=miauth.cli:main']
    },
    install_requires=[
        'cryptography'
    ],
    extras_require={
        'cli': [
            'bluepy==1.3.0'
        ]
    },
    python_requires=">=3.6",
)
